import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="aspire",
    version="0.11.0",
    data_files=[
        ("", ["src/aspire/config_default.yaml"]),
        ("", ["src/aspire/logging.conf"]),
    ],
    include_package_data=True,
    description="Algorithms for Single Particle Reconstruction",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://github.com/ComputationalCryoEM/ASPIRE-Python",
    author="Joakim Anden, Ayelet Heimowitz, Vineet Bansal, Robbie Brook, Itay Sason, Yoel Shkolnisky, Garrett Wright, Junchao Xia",
    author_email="devs.aspire@gmail.com",
    install_requires=[
        "click",
        "confuse>=2.0.0",
        "finufft",
        "gemmi>=0.4.8",
        "grpcio<=1.48.2",
        "joblib",
        "matplotlib>=3.2.0",
        "mrcfile",
        "numpy>=1.21.5",
        "packaging",
        "psutil",
        "pyfftw",
        "PyWavelets",
        "pillow",
        "ray",
        "scipy>=1.7.3",
        "scikit-learn",
        "scikit-image",
        "setuptools>=0.41",
        "tqdm",
    ],
    # Here we can call out specific extras,
    #   for example gpu packages which may not install for all users,
    #   or developer tools that are handy but not required for users.
    extras_require={
        "gpu_102": ["pycuda", "cupy-cuda102", "cufinufft==1.2"],
        "gpu_110": ["pycuda", "cupy-cuda110", "cufinufft==1.2"],
        "gpu_111": ["pycuda", "cupy-cuda111", "cufinufft==1.2"],
        "gpu_11x": ["pycuda", "cupy-cuda11x", "cufinufft==1.2"],
        "dev": [
            "black",
            "bumpversion",
            "check-manifest",
            "flake8>=3.7.0",
            "isort",
            "jupyter",
            "pyflakes",
            "pydocstyle",
            "parameterized",
            "pytest",
            "pytest-cov",
            "pytest-random-order",
            "pytest-xdist",
            "requests",
            "sphinxcontrib-bibtex",
            "sphinxcontrib-mermaid",
            "sphinx-gallery",
            "sphinx-rtd-theme>=0.4.2",
            "snakeviz",
            "tox",
            "twine",
        ],
    },
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    package_data={"aspire": ["config_default.yaml"]},
    zip_safe=True,
    test_suite="tests",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    entry_points={
        "console_scripts": [
            "aspire = aspire.__main__:main_entry",
        ]
    },
)



import logging

import numpy as np

from aspire.basis import Basis
from aspire.nufft import anufft, nufft
from aspire.utils import complex_type

logger = logging.getLogger(__name__)


class PolarBasis2D(Basis):
    """
    Define a derived class for polar Fourier representation for 2D images
    """

    def __init__(self, size, nrad=None, ntheta=None, dtype=np.float32):
        """
        Initialize an object for the 2D polar Fourier grid class

        :param size: The shape of the vectors for which to define the grid.
            May be a 2-tuple or an integer, in which case a square basis is assumed.
            Currently only square images are supported.
        :param nrad: The number of points in the radial dimension. Default is resolution // 2.
        :param ntheta: The number of points in the angular dimension. Default is 8 * nrad.
        """
        if isinstance(size, int):
            size = (size, size)
        ndim = len(size)
        assert ndim == 2, "Only two-dimensional grids are supported."
        assert len(set(size)) == 1, "Only square domains are supported."

        self.nrad = nrad
        self.ntheta = ntheta

        super().__init__(size, dtype=dtype)

        # this basis has complex coefficients
        self.coefficient_dtype = complex_type(self.dtype)

    def _build(self):
        """
        Build the internal data structure to 2D polar Fourier grid
        """
        logger.info("Represent 2D image in a polar Fourier grid")

        if self.nrad is None:
            self.nrad = self.nres // 2

        if self.ntheta is None:
            # try to use the same number as Fast FB basis
            self.ntheta = 8 * self.nrad

        if self.ntheta % 2 == 1:
            msg = "Only even values for ntheta are supported."
            logger.error(msg)
            raise NotImplementedError(msg)

        self.count = self.nrad * self.ntheta
        self._sz_prod = self.sz[0] * self.sz[1]

        # precompute the basis functions in 2D grids
        self.freqs = self._precomp()

    def _precomp(self):
        """
        Precomute the polar Fourier grid
        """
        omega0 = 2 * np.pi / (2 * self.nrad - 1)
        dtheta = 2 * np.pi / self.ntheta

        # only need half size of ntheta
        freqs = np.zeros((2, self.nrad * self.ntheta // 2), dtype=self.dtype)
        for i in range(self.ntheta // 2):
            freqs[0, i * self.nrad : (i + 1) * self.nrad] = np.arange(
                self.nrad
            ) * np.cos(i * dtheta)
            freqs[1, i * self.nrad : (i + 1) * self.nrad] = np.arange(
                self.nrad
            ) * np.sin(i * dtheta)

        freqs *= omega0
        return freqs

    def _evaluate(self, v):
        """
        Evaluate coefficients in standard 2D coordinate basis from those in polar Fourier basis

        :param v: A coefficient vector (or an array of coefficient vectors)
            in polar Fourier basis to be evaluated. The last dimension must equal to
            `self.count`.
        :return x: Image instance in standard 2D coordinate basis with
            resolution of `self.sz`.
        """
        v = v.reshape(-1, self.ntheta, self.nrad)

        nimgs = v.shape[0]

        half_size = self.ntheta // 2

        v = v[:, :half_size, :] + v[:, half_size:, :].conj()

        v = v.reshape(nimgs, self.nrad * half_size)

        x = anufft(v, self.freqs, self.sz, real=True)

        return x

    def _evaluate_t(self, x):
        """
        Evaluate coefficient in polar Fourier grid from those in standard 2D coordinate basis

        :param x: The Image instance representing coefficient array in the
            standard 2D coordinate basis to be evaluated.
        :return: The evaluation of the coefficient array `x` in the polar
            Fourier grid. This is an array of vectors whose first dimension
            corresponds to `x.shape[0]`, and last dimension equals `self.count`.
        """
        nimgs = x.shape[0]

        half_size = self.ntheta // 2

        pf = nufft(x, self.freqs)

        pf = pf.reshape((nimgs, self.nrad, half_size))
        v = np.concatenate((pf, pf.conj()), axis=1)

        # return v coefficients with the last dimension size of self.count
        v = v.reshape(nimgs, -1)
        return v



import numpy as np


class Numpy:
    asnumpy = staticmethod(lambda x: x)

    def __getattr__(self, item):
        """
        Catch-all method to to allow a straight pass-through \
        of any attribute that is not supported above.
        """

        return getattr(np, item)



"""
Created on Sep 10, 2019
@author: Ayelet Heimowitz, Amit Moscovich

Integrated into ASPIRE by Garrett Wright Feb 2021.
"""

import logging
import os
from collections import OrderedDict

import mrcfile
import numpy as np
from numpy import linalg as npla
from scipy.optimize import linprog
from scipy.signal.windows import dpss

from aspire.basis.ffb_2d import FFBBasis2D
from aspire.image import Image
from aspire.numeric import fft
from aspire.storage import StarFile
from aspire.utils import abs2, complex_type, grid_1d, grid_2d, voltage_to_wavelength

logger = logging.getLogger(__name__)


class CtfEstimator:
    """
    CtfEstimator Class ...
    """

    def __init__(
        self,
        pixel_size,
        cs,
        amplitude_contrast,
        voltage,
        psd_size,
        num_tapers,
        dtype=np.float32,
    ):
        """
        Instantiate a CtfEstimator instance.

        :param pixel_size: Size of the pixel in \u212b (Angstrom).
        :param cs: Spherical aberration in mm.
        :param amplitude_contrast: Amplitude contrast.
        :param voltage: Voltage of electron microscope.
        :param psd_size: Block size (in pixels) for PSD estimation.
        :param num_tapers: Number of tapers to apply in PSD estimation.
        :returns: CtfEstimator instance.
        """

        self.pixel_size = pixel_size
        self.cs = cs
        self.amplitude_contrast = amplitude_contrast
        self.voltage = voltage
        self.psd_size = psd_size
        self.num_tapers = num_tapers
        self.lmbd = voltage_to_wavelength(voltage) / 10.0  # Convert Angstrom to nm
        self.dtype = np.dtype(dtype)

        grid = grid_2d(psd_size, normalized=True, indexing="yx", dtype=self.dtype)

        # Note range is -half to half.
        self.r_ctf = grid["r"] / 2 * (10 / pixel_size)  # units: inverse nm

        self.theta = grid["phi"]
        self.defocus1 = 0
        self.defocus2 = 0
        self.angle = 0  # Radians
        self.h = 0

    def set_df1(self, df):
        """
        Sets defocus.

        :param df: Defocus value in the direction perpendicular to df2.
        """

        self.defocus1 = df

    def set_df2(self, df):
        """
        Sets defocus.

        :param df: Defocus value in the direction perpendicular to df1.
        """

        self.defocus2 = df

    def set_angle(self, angle):
        """
        Sets angle.

        :param angle: Angle (in Radians) between df1 and the x-axis.
        """

        self.angle = angle

    def generate_ctf(self):
        """
        Generates internal representation of the Contrast Transfer Function using parameters from this instance.
        """

        astigmatism_angle = np.full(
            shape=self.theta.shape, fill_value=self.angle, dtype=self.dtype
        )

        defocus_sum = np.full(
            shape=self.theta.shape,
            fill_value=self.defocus1 + self.defocus2,
            dtype=self.dtype,
        )

        defocus = defocus_sum + (
            (self.defocus1 - self.defocus2)
            * np.cos(2 * (self.theta - astigmatism_angle))
        )
        defocus_factor = np.pi * self.lmbd * self.r_ctf * defocus / 2
        amplitude_contrast_term = self.amplitude_contrast / np.sqrt(
            1 - self.amplitude_contrast**2
        )

        chi = (
            defocus_factor
            - np.pi * self.lmbd**3 * self.cs * 1e6 * self.r_ctf**2 / 2
            + amplitude_contrast_term
        )
        h = -np.sin(chi)
        self.h = h

    def micrograph_to_blocks(self, micrograph, block_size):
        """
        Preprocess micrograph into blocks using block_size.

        :param micrograph: Micrograph as NumPy array. #NOTE looks like F order
        :param blocksize: Size of the square blocks to partition micrograph.
        :return: NumPy array of blocks extracted from the micrograph.
        """

        # verify block_size is even
        assert block_size % 2 == 0

        if micrograph.ndim == 3:
            assert (
                micrograph.shape[0] == 1
            ), f"micrograph should be 2D or stack of 1 2D image: {micrograph.shape}"
            micrograph = micrograph[0]

        size_x = micrograph.shape[-1]
        size_y = micrograph.shape[-2]

        step_size = block_size // 2
        range_y = size_y // step_size - 1
        range_x = size_x // step_size - 1

        block_list = [
            micrograph[
                j * step_size : (j + 2) * step_size, i * step_size : (i + 2) * step_size
            ]
            for j in range(range_y)
            for i in range(range_x)
        ]
        blocks = np.asarray(block_list, dtype=micrograph.dtype)

        return blocks

    def normalize_blocks(self, blocks):
        """
        Preprocess CTF of micrograph using block_size.

        :param blocks: NumPy array of blocks extracted from the micrograph.
        :return: NumPy array of normalized blocks.
        """

        # Take block size from blocks
        block_size = blocks.shape[1]
        assert block_size == blocks.shape[2]

        # Create a sum and reshape so it may be broadcast with `block`.
        blocks_sum = np.sum(blocks, axis=(-1, -2))[:, np.newaxis, np.newaxis]

        blocks -= blocks_sum / (block_size**2)

        return blocks

    def preprocess_micrograph(self, micrograph, block_size):
        """
        Preprocess micrograph into normalized blocks using block_size.

        :param micrograph: Micrograph as NumPy array. #NOTE looks like F order
        :param blocksize: Size of the square blocks to partition micrograph.
        :return: NumPy array of normalized blocks extracted from the micrograph.
        """

        return self.normalize_blocks(self.micrograph_to_blocks(micrograph, block_size))

    def tapers(self, N, NW, L):
        """
        Compute data tapers (which are discrete prolate spheroidal sequences (dpss))

        Uses scipy implementation, see:
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.dpss.html

        :param N: Size of each taper
        :param NW: Half Bandwidth
        :param L: Number of tapers
        :return: NumPy array of data tapers
        """

        # Note the original ASPIRE implementation is negated from original scipy...
        #  but at time of writing subsequent code was agnostic to sign.
        return dpss(M=N, NW=NW, Kmax=L, return_ratios=False).T

    def estimate_psd(self, blocks, tapers_1d):
        """
        Estimate the power spectrum of the micrograph using the multi-taper method

        :param blocks: 3-D NumPy array containing windows extracted from the micrograph in the preprocess function.
        :param tapers_1d: NumPy array of data tapers.
        :return: NumPy array of estimated power spectrum.
        """

        num_1d_tapers = tapers_1d.shape[-1]
        tapers_1d = tapers_1d.astype(complex_type(self.dtype), copy=False)

        blocks_mt = np.zeros(blocks[0, :, :].shape, dtype=self.dtype)

        blocks_tapered = np.zeros(blocks[0, :, :].shape, dtype=complex_type(self.dtype))

        taper_2d = np.zeros(
            (blocks.shape[1], blocks.shape[2]), dtype=complex_type(self.dtype)
        )

        for ax1 in range(num_1d_tapers):
            for ax2 in range(num_1d_tapers):
                np.matmul(
                    tapers_1d[:, ax1, np.newaxis],
                    tapers_1d[:, ax2, np.newaxis].T,
                    out=taper_2d,
                )

                for m in range(blocks.shape[0]):
                    np.multiply(blocks[m, :, :], taper_2d, out=blocks_tapered)
                    blocks_mt_post_fft = fft.fftn(blocks_tapered, axes=(-2, -1))
                    blocks_mt += abs2(blocks_mt_post_fft)

        blocks_mt /= blocks.shape[0] ** 2
        blocks_mt /= tapers_1d.shape[0] ** 2

        amplitude_spectrum = fft.fftshift(
            blocks_mt
        )  # max difference 10^-13, max relative difference 10^-14

        return Image(amplitude_spectrum)

    def elliptical_average(self, ffbbasis, amplitude_spectrum, circular):
        """
        Computes radial/elliptical average of the power spectrum

        :param ffbbasis: FFBBasis instance.
        :param amplitude_spectrum: Power spectrum.
        :param circular: True for radial averaging and False for elliptical averaging.
        :return: PSD and noise as 2-tuple of NumPy arrays.
        """

        # RCOPT, come back and change the indices for this method
        coeffs_s = ffbbasis.evaluate_t(amplitude_spectrum).T
        coeffs_n = coeffs_s.copy()

        coeffs_s[np.argwhere(ffbbasis._indices["ells"] == 1)] = 0
        if circular:
            coeffs_s[np.argwhere(ffbbasis._indices["ells"] == 2)] = 0
            noise = amplitude_spectrum
        else:
            coeffs_n[np.argwhere(ffbbasis._indices["ells"] == 0)] = 0
            coeffs_n[np.argwhere(ffbbasis._indices["ells"] == 2)] = 0
            noise = ffbbasis.evaluate(coeffs_n.T)

        psd = ffbbasis.evaluate(coeffs_s.T)

        return psd, noise

    def background_subtract_1d(
        self, amplitude_spectrum, linprog_method="highs", n_low_freq_cutoffs=14
    ):
        """
        Estimate and subtract the background from the power spectrum

        :param amplitude_spectrum: Estimated power spectrum
        :param linprog_method: Method passed to linear program solver (scipy.optimize.linprog).
        :param n_low_freq_cutoffs: Low frequency cutoffs (loop iterations).
        :return: 2-tuple of NumPy arrays (PSD after noise subtraction and estimated noise)
        """
        # compute radial average
        center = amplitude_spectrum.shape[-1] // 2

        if amplitude_spectrum.ndim == 3:
            if amplitude_spectrum.shape[0] != 1:
                raise ValueError(
                    f"Invalid dimension 0 for amplitude_spectrum {amplitude_spectrum.shape}"
                )
            amplitude_spectrum = amplitude_spectrum[0]
        elif amplitude_spectrum.ndim > 3:
            raise ValueError(
                f"Invalid ndimension for amplitude_spectrum {amplitude_spectrum.shape}"
            )

        amplitude_spectrum = amplitude_spectrum.asnumpy()[0, center, center:]
        amplitude_spectrum = amplitude_spectrum[
            0 : 3 * amplitude_spectrum.shape[-1] // 4
        ]

        final_signal = np.zeros(
            (n_low_freq_cutoffs - 1, amplitude_spectrum.shape[-1]), dtype=self.dtype
        )
        final_background = np.ones(
            (n_low_freq_cutoffs - 1, amplitude_spectrum.shape[-1]), dtype=self.dtype
        )

        for low_freq_cutoff in range(1, n_low_freq_cutoffs):
            signal = amplitude_spectrum[low_freq_cutoff:]
            signal = np.ravel(signal)
            N = amplitude_spectrum.shape[-1] - low_freq_cutoff

            f = np.concatenate((np.ones(N), -1 * np.ones(N)), axis=0)

            superposition_condition = np.concatenate(
                (-1 * np.eye(N), np.eye(N)), axis=1
            )

            monotone_condition = np.diag(np.full((N - 1), -1), -1) + np.diag(
                np.ones(N), 0
            )
            monotone_condition = monotone_condition[1:]

            convex_condition = (
                np.diag(np.full((N - 1), -1), -1)
                + np.diag(np.full(N, 2), 0)
                + np.diag(np.full((N - 1), -1), 1)
            )
            convex_condition = np.concatenate(
                (np.zeros((N, N)), convex_condition), axis=1
            )
            convex_condition = np.roll(convex_condition, -1, axis=0)
            convex_condition[N - 2 :] = 0

            positivity_condition = np.concatenate(
                (np.zeros((N, N)), -1 * np.eye(N)), axis=1
            )

            A = np.concatenate(
                (superposition_condition, convex_condition, positivity_condition),
                axis=0,
            )

            # The original code used `bounds`,
            #   but for many problems, linprog reports infeasable constraints.
            # In practice for a micrograph from the paper, and our tutorial,
            #   the code seems to work better without it...
            # ASPIRE #417

            x = linprog(
                f,
                A_ub=A,
                b_ub=np.zeros(A.shape[0]),
                method=linprog_method,
            )

            if not x.success:
                raise RuntimeError("Linear program did not succeed. Halting")

            background = x.x[N:]

            bs_psd = signal - background

            final_signal[low_freq_cutoff - 1, low_freq_cutoff:] = bs_psd.T
            # expected difference: 10^-7 (absolute)
            final_background[low_freq_cutoff - 1, low_freq_cutoff:] = background.T

        return final_signal, final_background

    def opt1d(
        self,
        amplitude_spectrum,
        pixel_size,
        cs,
        lmbd,
        w,
        N,
        min_defocus=500,
        max_defocus=10000,
    ):
        """
        Find optimal defocus for the radially symmetric case (where no astigmatism is present)

        :param amplitude_spectrum: Estimated power specrtum.
        :param pixel_size: Pixel size in \u212b (Angstrom).
        :param cs: Spherical aberration in mm.
        :param lmbd: Electron wavelength \u212b (Angstrom).
        :param w: Amplitude contrast.
        :param N: Number of rows (or columns) in the estimate power spectrum.
        :param min_defocus: Start of defocus loop scan.
        :param max_defocus: End of defocus loop scan.
        :return: 2-tuple of NumPy arrays (Estimated average of defocus and low_freq_cutoff)
        """

        center = N // 2

        grid = grid_1d(N, normalized=True, dtype=self.dtype)
        rb = grid["r"][center:] / 2

        r_ctf = rb * (10 / pixel_size)  # units: inverse nm

        signal = amplitude_spectrum.T
        signal = np.maximum(0.0, signal)
        signal = np.sqrt(signal)
        signal = signal[: 3 * signal.shape[0] // 4]

        r_ctf_sq = r_ctf**2
        c = np.zeros((max_defocus - min_defocus, signal.shape[1]), dtype=self.dtype)

        for f in range(min_defocus, max_defocus):
            ctf_im = np.abs(
                np.sin(
                    np.pi * lmbd * f * r_ctf_sq
                    - 0.5 * np.pi * (lmbd**3) * cs * 1e6 * r_ctf_sq**2
                    + w
                )
            )
            ctf_im = ctf_im[: signal.shape[0]]

            ctf_im = np.reshape(ctf_im, (ctf_im.shape[0], 1))
            ctf_im = np.tile(ctf_im, (1, signal.shape[1]))

            for m in range(0, signal.shape[1]):
                signal[:, m] = signal[:, m] - np.mean(signal[m + 1 :, m], axis=0)
                ctf_im[:, m] = ctf_im[:, m] - np.mean(ctf_im[m + 1 :, m], axis=0)
                ctf_im[: m + 1, m] = np.zeros((m + 1))
                signal[: m + 1, m] = np.zeros((m + 1))

            Sx = np.sqrt(np.sum(ctf_im**2, axis=0))
            Sy = np.sqrt(np.sum(signal**2, axis=0))
            c[f - min_defocus, :] = np.sum(ctf_im * signal, axis=0) / (Sx * Sy)

        avg_defocus, low_freq_cutoff = np.unravel_index(np.argmax(c), c.shape)[:2]
        avg_defocus += min_defocus

        return avg_defocus, low_freq_cutoff

    def background_subtract_2d(self, signal, background_p1, max_col):
        """
        Subtract background from estimated power spectrum

        :param signal: Estimated power spectrum
        :param background_p1: 1-D background estimation
        :param max_col: Internal variable, returned as the second parameter from opt1d.
        :return: 2-tuple of NumPy arrays (Estimated PSD without noise and estimated noise).
        """

        signal = signal.asnumpy()

        N = signal.shape[1]
        grid = grid_2d(N, normalized=False, indexing="yx", dtype=self.dtype)

        radii = np.sqrt((grid["x"] / 2) ** 2 + (grid["y"] / 2) ** 2)

        background = np.zeros(signal.shape, dtype=self.dtype)
        for r in range(max_col + 2, background_p1.shape[1]):
            background[:, (r < radii) & (radii <= r + 1)] = background_p1[max_col, r]
        mask = radii <= max_col + 2
        background[:, mask] = signal[:, mask]

        signal = signal - background
        signal = np.maximum(0, signal)

        return Image(signal), Image(background)

    def pca(self, signal, pixel_size, g_min, g_max):
        """

        :param signal: Estimated power spectrum.
        :param pixel_size: Pixel size in \u212b (Angstrom).
        :param g_min: Inverse of minimun resolution for PSD.
        :param g_max: Inverse of maximum resolution for PSD.
        :return: ratio.
        """

        # RCOPT
        signal = signal.asnumpy()[0].T

        N = signal.shape[0]
        center = N // 2

        grid = grid_2d(N, normalized=True, indexing="yx", dtype=self.dtype)

        r_ctf = grid["r"] / 2 * (10 / pixel_size)

        grid = grid_2d(N, normalized=False, indexing="yx", dtype=self.dtype)
        X = grid["x"]
        Y = grid["y"]

        signal = signal - np.min(signal)

        rad_sq_min = N * pixel_size / g_min
        rad_sq_max = N * pixel_size / g_max

        min_limit = r_ctf[center, (center + np.floor(rad_sq_min)).astype(int)]
        signal[r_ctf < min_limit] = 0

        max_limit = r_ctf[center, (center + np.ceil(rad_sq_max)).astype(int)]
        signal = np.where(r_ctf > max_limit, 0, signal)

        moment_02 = Y**2 * signal
        moment_02 = np.sum(moment_02, axis=(0, 1))

        moment_11 = Y * X * signal
        moment_11 = np.sum(moment_11, axis=(0, 1))

        moment_20 = X**2 * signal
        moment_20 = np.sum(moment_20, axis=(0, 1))

        moment_mat = np.zeros((2, 2))
        moment_mat[0, 0] = moment_20
        moment_mat[1, 1] = moment_02
        moment_mat[0, 1] = moment_11
        moment_mat[1, 0] = moment_11

        moment_evals = npla.eigvalsh(moment_mat)
        ratio = moment_evals[0] / moment_evals[1]

        return ratio

    def gd(
        self,
        signal,
        df1,
        df2,
        angle_ast,
        r,
        theta,
        pixel_size,
        g_min,
        g_max,
        amplitude_contrast,
        lmbd,
        cs,
    ):
        """
        Runs gradient ascent to optimize defocus parameters

        :param signal: Estimated power spectrum
        :param df1: Defocus value in the direction perpendicular to df2.
        :param df2: Defocus value in the direction perpendicular to df1.
        :param angle_ast: Angle between df1 and the x-axis, Radians.
        :param r: Magnitude of spatial frequencies.
        :param theta: Phase of spatial frequencies.
        :param pixel_size: Pixel size in \u212b (Angstrom).
        :param g_min: Inverse of minimun resolution for PSD.
        :param g_max: Inverse of maximum resolution for PSD.
        :param amplitude_contrast: Amplitude contrast.
        :param lmbd: Electron wavelength \u212b (Angstrom).
        :param cs: Spherical aberration in mm.
        :return: Optimal defocus parameters
        """

        # step size
        alpha1 = 1e5
        alpha2 = 1e4

        # initialization
        x = df1 + df2
        y = (df1 - df2) * np.cos(2 * angle_ast)
        z = (df1 - df2) * np.sin(2 * angle_ast)

        a = np.pi * lmbd * r**2 / 2
        b = np.pi * lmbd**3 * cs * 1e6 * r**4 / 2 - np.full(
            shape=r.shape, fill_value=amplitude_contrast, dtype=self.dtype
        )

        signal = signal.asnumpy()[0].T
        N = signal.shape[1]
        center = N // 2

        rad_sq_min = N * pixel_size / g_min
        rad_sq_max = N * pixel_size / g_max

        max_val = r[center, int(center - 1 + np.floor(rad_sq_max))]
        min_val = r[center, int(center - 1 + np.ceil(rad_sq_min))]

        mask = (r <= max_val) & (r > min_val)
        a = a[mask]
        b = b[mask]
        signal = signal[..., mask]
        r = r[mask]
        theta = theta[mask]

        sum_A = np.sum(signal**2)

        dx = 1
        dy = 1
        dz = 1

        stop_cond = 1e-20
        iter_no = 1

        while np.maximum(np.maximum(dx, dy), dz) > stop_cond:
            inner_cosine = y * np.cos(2 * theta) + z * np.sin(2 * theta)
            psi = a * x + a * inner_cosine - b
            outer_sine = np.sin(psi)
            outer_cosine = np.cos(psi)

            sine_x_term = a
            sine_y_term = a * np.cos(2 * theta)
            sine_z_term = a * np.sin(2 * theta)

            c1 = np.sum(np.abs(outer_sine) * signal)
            c2 = np.sqrt(sum_A * np.sum(outer_sine**2))

            # gradients of numerator
            dx_c1 = np.sum(np.sign(outer_sine) * outer_cosine * a * signal)

            dy_c1 = np.sum(
                np.sign(outer_sine) * outer_cosine * a * np.cos(2 * theta) * signal
            )
            dz_c1 = np.sum(
                np.sign(outer_sine) * outer_cosine * a * np.sin(2 * theta) * signal
            )

            derivative_sqrt = 1 / (2 * np.sqrt(sum_A * np.sum(outer_sine**2)))

            derivative_sine2 = 2 * outer_sine * outer_cosine

            #  gradients of denomenator
            dx_c2 = derivative_sqrt * sum_A * np.sum(derivative_sine2 * sine_x_term)
            dy_c2 = derivative_sqrt * sum_A * np.sum(derivative_sine2 * sine_y_term)
            dz_c2 = derivative_sqrt * sum_A * np.sum(derivative_sine2 * sine_z_term)

            # gradients
            dx = (dx_c1 * c2 - dx_c2 * c1) / c2**2
            dy = (dy_c1 * c2 - dy_c2 * c1) / c2**2
            dz = (dz_c1 * c2 - dz_c2 * c1) / c2**2

            # update
            x = x + alpha1 * dx
            y = y + alpha2 * dy
            z = z + alpha2 * dz

            if iter_no < 2:
                stop_cond = np.minimum(np.minimum(dx, dy), dz) / 1000

            if iter_no > 400:
                stop_cond = np.maximum(np.maximum(dx, dy), dz) + 1

            iter_no = iter_no + 1

        df1 = (x + np.abs(y + z * 1j)) / 2
        df2 = (x - np.abs(y + z * 1j)) / 2
        angle_ast = np.angle(y + z * 1j) / 2  # Radians

        inner_cosine = y * np.cos(2 * theta) + z * np.sin(2 * theta)
        outer_sine = np.sin(a * x + a * inner_cosine - b)
        outer_cosine = np.cos(a * x + a * inner_cosine - b)

        sine_x_term = a
        sine_y_term = a * np.cos(2 * theta)
        sine_z_term = a * np.sin(2 * theta)

        c1 = np.sum(np.abs(outer_sine) * signal)
        c2 = np.sqrt(sum_A * np.sum(outer_sine**2))

        p = c1 / c2

        return df1, df2, angle_ast, p

    # Note, This doesn't actually use anything from the class.
    # It is used in a solver loop of some sort, so it may not be correct
    # to just use what is avail in the obj.
    def write_star(self, name, params_dict, output_dir):
        """
        Writes CTF parameters to starfile for a single micrograph.
        """
        data_block = {}
        data_block["_rlnMicrographName"] = name
        data_block["_rlnDefocusU"] = params_dict["defocus_u"]  # Should already be in A
        data_block["_rlnDefocusV"] = params_dict["defocus_v"]  # Should already be in A
        data_block["_rlnDefocusAngle"] = (
            params_dict["defocus_ang"] * 180 / np.pi
        )  # Convert from radian to degree
        data_block["_rlnSphericalAberration"] = params_dict["cs"]
        data_block["_rlnAmplitudeContrast"] = params_dict["amplitude_contrast"]
        data_block["_rlnVoltage"] = params_dict["voltage"]
        data_block["_rlnMicrographPixelSize"] = params_dict["pixel_size"]
        blocks = OrderedDict()
        blocks["root"] = data_block
        star = StarFile(blocks=blocks)
        star.write(os.path.join(output_dir, os.path.splitext(name)[0]) + ".star")


def estimate_ctf(
    data_folder,
    pixel_size=1.0,
    cs=2.0,
    amplitude_contrast=0.07,
    voltage=300,
    num_tapers=2,
    psd_size=512,
    g_min=30,
    g_max=5,
    output_dir="results",
    dtype=np.float32,
    save_ctf_images=False,
    save_noise_images=False,
):
    """
    Given paramaters estimates CTF from experimental data
    and returns CTF as a mrc file.
    """

    dtype = np.dtype(dtype)
    assert dtype in (np.float32, np.float64)

    dir_content = os.scandir(data_folder)

    mrc_files = [f.name for f in dir_content if os.path.splitext(f)[1] == ".mrc"]
    mrcs_files = [f.name for f in dir_content if os.path.splitext(f)[1] == ".mrcs"]
    file_names = mrc_files + mrcs_files

    amp = amplitude_contrast
    amplitude_contrast = np.arctan(
        amplitude_contrast / np.sqrt(1 - amplitude_contrast**2)
    )

    lmbd = voltage_to_wavelength(voltage) / 10  # Convert from Angstrom to nm

    ctf_object = CtfEstimator(
        pixel_size, cs, amplitude_contrast, voltage, psd_size, num_tapers, dtype=dtype
    )

    # Note for repro debugging, suggest use of doubles,
    #   closer to original code.
    ffbbasis = FFBBasis2D((psd_size, psd_size), 2, dtype=dtype)

    results = {}
    for name in file_names:
        with mrcfile.open(
            os.path.join(data_folder, name), mode="r", permissive=True
        ) as mrc:
            micrograph = mrc.data

        # Try to match dtype used in Basis instance
        micrograph = micrograph.astype(dtype, copy=False)

        micrograph_blocks = ctf_object.preprocess_micrograph(micrograph, psd_size)

        tapers_1d = ctf_object.tapers(psd_size, num_tapers / 2, num_tapers)

        signal_observed = ctf_object.estimate_psd(micrograph_blocks, tapers_1d)

        amplitude_spectrum, _ = ctf_object.elliptical_average(
            ffbbasis, signal_observed, True
        )  # absolute differenceL 10^-14. Relative error: 10^-7

        # Optionally changing to: linprog_method='simplex',
        # will more deterministically repro results in exchange for speed.
        # linprog_method was changed from 'interior-point' to 'highs' due to
        # "interior-point' being deprecated.
        signal_1d, background_1d = ctf_object.background_subtract_1d(
            amplitude_spectrum, linprog_method="highs"
        )

        avg_defocus, low_freq_skip = ctf_object.opt1d(
            signal_1d,
            pixel_size,
            cs,
            lmbd,  # (Angstrom)
            amplitude_contrast,
            signal_observed.shape[-1],
        )

        low_freq_skip = 12
        signal, background_2d = ctf_object.background_subtract_2d(
            signal_observed, background_1d, low_freq_skip
        )

        ratio = ctf_object.pca(signal_observed, pixel_size, g_min, g_max)

        signal, additional_background = ctf_object.elliptical_average(
            ffbbasis, signal.sqrt(), False
        )

        background_2d = background_2d + additional_background

        initial_df1 = (avg_defocus * 2) / (1 + ratio)
        initial_df2 = (avg_defocus * 2) - initial_df1

        grid = grid_2d(psd_size, normalized=True, indexing="yx", dtype=dtype)

        r_ctf = grid["r"] / 2 * (10 / pixel_size)
        theta = grid["phi"]

        angle = -5 / 12 * np.pi  # Radians (-75 degrees)
        cc_array = np.zeros((6, 4))
        for a in range(0, 6):
            df1, df2, angle_ast, p = ctf_object.gd(
                signal,
                initial_df1,
                initial_df2,
                angle + a * np.pi / 6.0,  # Radians, + a*30degrees
                r_ctf,
                theta,
                pixel_size,
                g_min,
                g_max,
                amplitude_contrast,
                lmbd,  # (Angstrom)
                cs,
            )

            cc_array[a, 0] = df1
            cc_array[a, 1] = df2
            cc_array[a, 2] = angle_ast  # Radians
            cc_array[a, 3] = p
        ml = np.argmax(cc_array[:, 3], -1)

        result = {
            "defocus_u": cc_array[ml, 0] * 10,  # Convert from nm to A
            "defocus_v": cc_array[ml, 1] * 10,  # Convert from nm to A
            "defocus_ang": cc_array[ml, 2],  # Radians
            "cs": cs,
            "voltage": voltage,
            "pixel_size": pixel_size,
            "amplitude_contrast": amp,
        }
        results[name] = result

        # we write each micrograph's ctf parameters to an individual starfile
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)
        ctf_object.write_star(name, result, output_dir)

        if save_noise_images:
            with mrcfile.new(
                os.path.join(output_dir, os.path.splitext(name)[0] + "_noise.mrc"),
                overwrite=True,
            ) as mrc:
                mrc.set_data(background_2d.asnumpy()[0].astype(np.float32))
                mrc.voxel_size = pixel_size

        if save_ctf_images:
            ctf_object.set_df1(cc_array[ml, 0])  # Angstrom
            ctf_object.set_df2(cc_array[ml, 1])  # Angstrom
            ctf_object.set_angle(cc_array[ml, 2])  # Radians
            ctf_object.generate_ctf()
            df = (cc_array[ml, 0] + cc_array[ml, 1]) * np.ones(
                theta.shape, theta.dtype
            ) + (cc_array[ml, 0] - cc_array[ml, 1]) * np.cos(
                2 * theta - 2 * cc_array[ml, 2] * np.ones(theta.shape, theta.dtype)
            )
            ctf_im = -np.sin(
                np.pi * lmbd * r_ctf**2 / 2 * (df - lmbd**2 * r_ctf**2 * cs * 1e6)
                + amplitude_contrast
            )
            ctf_signal = np.zeros(ctf_im.shape, ctf_im.dtype)
            ctf_signal[: ctf_im.shape[0] // 2, :] = ctf_im[: ctf_im.shape[0] // 2, :]
            ctf_signal[ctf_im.shape[0] // 2 + 1 :, :] = signal.asnumpy()[
                :, :, ctf_im.shape[0] // 2 + 1
            ]

            with mrcfile.new(
                os.path.join(output_dir, os.path.splitext(name)[0] + ".ctf"),
                overwrite=True,
            ) as mrc:
                mrc.set_data(np.float32(ctf_signal))
                mrc.voxel_size = pixel_size

    return results



import cupy as cp


class Cupy:
    def __getattr__(self, item):
        """
        Catch-all method to to allow a straight pass-through of any attribute that is not supported above.
        """
        return getattr(cp, item)



import inspect
import logging

import numpy as np
from scipy.interpolate import RegularGridInterpolator

from aspire.utils import grid_2d, voltage_to_wavelength
from aspire.utils.filter_to_fb_mat import filter_to_fb_mat

logger = logging.getLogger(__name__)


def evaluate_src_filters_on_grid(src):
    """
    Given an ImageSource object, compute the source's unique filters
    at the filter_indices specified in its metadata.

    :return: an `src.L x src.L x len(src.filter_indices)`
        array containing the evaluated filters at each gridpoint
    """

    grid2d = grid_2d(src.L, indexing="yx", dtype=src.dtype)
    omega = np.pi * np.vstack((grid2d["x"].flatten(), grid2d["y"].flatten()))

    h = np.empty((omega.shape[-1], len(src.filter_indices)), dtype=src.dtype)
    for i, filt in enumerate(src.unique_filters):
        idx_k = np.where(src.filter_indices == i)[0]
        if len(idx_k) > 0:
            filter_values = filt.evaluate(omega)
            h[:, idx_k] = np.column_stack((filter_values,) * len(idx_k))

    h = np.reshape(h, grid2d["x"].shape + (len(src.filter_indices),))

    return h


# TODO: filters should probably be dtyped...
class Filter:
    def __init__(self, dim=None, radial=False):
        self.dim = dim
        self.radial = radial

    def __mul__(self, other):
        return MultiplicativeFilter(self, other)

    def __str__(self):
        """
        Show class name of Filter
        :return: A string of class name
        """
        return self.__class__.__name__

    def evaluate(self, omega):
        """
        Evaluate the filter at specified frequencies.

        :param omega: A vector of size n (for 1d filters), or an array of size 2-by-n, representing the spatial
            frequencies at which the filter is to be evaluated. These are normalized so that pi is equal to the Nyquist
            frequency.
        :return: The value of the filter at the specified frequencies.
        """
        if omega.ndim == 1:
            assert self.radial, "Cannot evaluate a non-radial filter on 1D input array."
        elif omega.ndim == 2 and self.dim:
            assert (
                omega.shape[0] == self.dim
            ), f"Omega must be of size {self.dim} x n; Passed omega.shape {omega.shape}"

        if self.radial:
            if omega.ndim > 1:
                omega = np.sqrt(np.sum(omega**2, axis=0))
            omega, idx = np.unique(omega, return_inverse=True)
            omega = np.vstack((omega, np.zeros_like(omega)))

        h = self._evaluate(omega)

        if self.radial:
            h = np.take(h, idx)

        return h

    def _evaluate(self, omega):
        raise NotImplementedError("Subclasses should implement this method")

    def fb_mat(self, fbasis):
        """
        Represent the filter in FB basis matrix
        """
        return filter_to_fb_mat(self.evaluate, fbasis)

    def scale(self, c=1):
        """
        Scale filter by a constant factor

        :param c: The scaling factor. For c < 1, it dilates the filter(s) in frequency, while for c > 1,
            it compresses (default 1).
        :return: A ScaledFilter object
        """
        return ScaledFilter(self, c)

    def evaluate_grid(self, L, dtype=np.float32, *args, **kwargs):
        """
        Generates a two dimensional grid with prescribed dtype,
        yielding the values (omega) which are then evaluated by
        the filter's evaluate method.

        Passes arbritrary args and kwargs down to self.evaluate method.

        :param L: Number of grid points (L by L).
        :param dtype: dtype of grid, defaults np.float32.
        :return: Filter values at omega's points.
        """

        # Note we can probably unwind the "F"/m_reshape here
        grid2d = grid_2d(L, indexing="yx", dtype=dtype)
        omega = np.pi * np.vstack((grid2d["x"].flatten(), grid2d["y"].flatten()))
        h = self.evaluate(omega, *args, **kwargs)

        h = h.reshape(grid2d["x"].shape)

        return h

    def dual(self):
        return DualFilter(self)

    @property
    def sign(self):
        """
        A Filter object to evaluate the signs of the underlying filter.
        """
        return LambdaFilter(self, np.sign)


class DualFilter(Filter):
    """
    A Filter object that is dual to origin one, namely g(w)=f(-w)
    """

    def __init__(self, filter_in):
        self._filter = filter_in
        super().__init__()

    def evaluate(self, omega):
        return self._filter.evaluate(-omega)


class FunctionFilter(Filter):
    """
    A Filter object that is instantiated directly using a 1D or 2D function, which is then directly used for evaluating
    the filter.
    """

    def __init__(self, f, dim=None):
        n_args = len(inspect.signature(f).parameters)
        assert n_args in (1, 2), "Only 1D or 2D functions are supported"

        assert dim in (None, 1, 2), "Only 1D or 2D dimensions are supported"
        dim = dim or n_args

        self.f = f  # will be used directly in this Filter's evaluate method
        # Note: The function may well be radial from the caller's perspective, but we won't be applying it in a radial
        # manner if the function we were initialized from expected 2 arguments
        # (i.e. at runtime, we will still expect the incoming omega values to have x and y components).
        super().__init__(dim=dim, radial=dim > n_args)

    def _evaluate(self, omega):
        return self.f(*omega)


class PowerFilter(Filter):
    """
    A Filter object that is composed of a regular `Filter` object, but evaluates it to a specified power.
    """

    def __init__(self, filter, power=1):
        self._filter = filter
        self._power = power
        super().__init__(dim=filter.dim, radial=filter.radial)

    def _evaluate(self, omega):
        return self._filter.evaluate(omega) ** self._power

    def evaluate_grid(self, L, dtype=np.float32, *args, **kwargs):
        """
        Calls the provided filter's evaluate_grid method in case there is an optimization.

        If no optimized method is provided, falls back to base `evaluate_grid`.

        See `Filter.evaluate_grid` for usage.
        """

        return (
            self._filter.evaluate_grid(L, dtype=dtype, *args, **kwargs) ** self._power
        )


class LambdaFilter(Filter):
    """
    A Filter object to evaluate lambda function of a regular `Filter`.
    """

    def __init__(self, filter, f):
        self._filter = filter
        self._f = f
        super().__init__(dim=filter.dim, radial=filter.radial)

    def _evaluate(self, omega):
        return self._f(self._filter.evaluate(omega))


class MultiplicativeFilter(Filter):
    """
    A Filter object that returns the product of the evaluation of its individual filters
    """

    def __init__(self, *args):
        super().__init__(dim=args[0].dim, radial=all(c.radial for c in args))
        self._components = args

    def _evaluate(self, omega):
        res = 1
        for c in self._components:
            res *= c.evaluate(omega)
        return res


class ScaledFilter(Filter):
    """
    A Filter object that is composed of a regular `Filter` object, but evaluates it on a scaled omega.
    """

    def __init__(self, filt, scale):
        self._filter = filt
        self._scale = scale
        super().__init__(dim=filt.dim, radial=filt.radial)

    def _evaluate(self, omega):
        return self._filter.evaluate(omega / self._scale)

    def __str__(self):
        """
        Show class name of ScaledFilter and related information

        :return: A string of class name and related information
        """
        return f"ScaledFilter (scales {self._filter} by {self._scale})"


class ArrayFilter(Filter):
    def __init__(self, xfer_fn_array):
        """
        A Filter corresponding to the filter with the specified transfer function.

        :param xfer_fn_array: The transfer function of the filter in the form of an array of one or two dimensions.
        """
        dim = xfer_fn_array.ndim
        assert dim in (1, 2), "Only dimensions 1 and 2 supported."

        super().__init__(dim=dim, radial=False)

        # sz is assigned before we do anything with xfer_fn_array
        self.sz = xfer_fn_array.shape

        # The following code, though superficially different from the MATLAB code its copied from,
        # results in the same behavior.
        # TODO: This could use documentation - very unintuitive!
        if dim == 1:
            # If we have a vector of even length, then append the first element to the last
            if xfer_fn_array.shape[0] % 2 == 0:
                xfer_fn_array = np.concatenate(
                    (xfer_fn_array, np.array([xfer_fn_array[0]]))
                )
        elif dim == 2:
            # If we have a 2d array with an even number of rows, append the first row reversed at the bottom
            if xfer_fn_array.shape[0] % 2 == 0:
                xfer_fn_array = np.vstack((xfer_fn_array, xfer_fn_array[0, ::-1]))
            # If we have a 2d array with an even number of columns, append the first column reversed at the right
            if xfer_fn_array.shape[1] % 2 == 0:
                xfer_fn_array = np.hstack(
                    (xfer_fn_array, xfer_fn_array[::-1, 0][:, np.newaxis])
                )

        self.xfer_fn_array = xfer_fn_array

    def _evaluate(self, omega):
        _input_pts = tuple(np.linspace(1, x, x) for x in self.xfer_fn_array.shape)

        # TODO: This part could do with some documentation - not intuitive!
        temp = np.array(self.sz)[:, np.newaxis]
        omega = (omega / (2 * np.pi)) * temp
        omega += np.floor(temp / 2) + 1

        # Emulating the behavior of interpn(V,X1q,X2q,X3q,...) in MATLAB
        # The original MATLAB was using 'linear' and zero fill.
        # We will use 'linear' but fill_value=None which will extrapolate
        #  for values slightly outside the interpolation grid bounds.
        interpolator = RegularGridInterpolator(
            _input_pts,
            # https://github.com/scipy/scipy/issues/17718
            self.xfer_fn_array.astype(np.float64),
            method="linear",
            bounds_error=False,
            fill_value=None,
        )

        result = interpolator(
            # Split omega into input arrays and stack depth-wise because that's how
            # the interpolator wants it
            np.dstack(np.split(omega, len(self.sz)))
        )

        # Result is 1 x np.prod(self.sz) in shape; convert to a 1-d vector
        result = np.squeeze(result, 0)

        return result

    def evaluate_grid(self, L, dtype=np.float32, *args, **kwargs):
        """
        Optimized evaluate_grid method for ArrayFilter.

        If evaluate_grid is called with a resolution L that matches
        the transfer function `xfer_fn_array` resolution,
        we do not need to generate a grid, setup interpolation, and
        evaluate by interpolation. We can instead use the transfer
        function directly.

        In the case the grid is not a match, we fall back to the
        base `evaluate_grid` implementation.

        See Filter.evaluate_grid for usage.
        """

        if all(dim == L for dim in self.xfer_fn_array.shape):
            logger.debug(
                "Size of transfer function matches evaluate_grid size L exactly,"
                " skipping grid generation and interpolation."
            )
            res = self.xfer_fn_array
        else:
            # Otherwise call parent code to generate a grid then evaluate.
            res = super().evaluate_grid(L, dtype=dtype, *args, **kwargs)
        return res


class ScalarFilter(Filter):
    def __init__(self, dim=None, value=1):
        super().__init__(dim=dim, radial=True)
        self.value = value

    def __repr__(self):
        return f"Scalar Filter (dim={self.dim}, value={self.value})"

    def _evaluate(self, omega):
        return self.value * np.ones_like(omega)


class ZeroFilter(ScalarFilter):
    def __init__(self, dim=None):
        super().__init__(dim=dim, value=0)


class IdentityFilter(ScalarFilter):
    def __init__(self, dim=None):
        super().__init__(dim=dim, value=1)


class CTFFilter(Filter):
    def __init__(
        self,
        pixel_size=10,
        voltage=200,
        defocus_u=15000,
        defocus_v=15000,
        defocus_ang=0,
        Cs=2.26,
        alpha=0.07,
        B=0,
    ):
        """
        A CTF (Contrast Transfer Function) Filter

        :param pixel_size:  Pixel size in angstrom
        :param voltage:     Electron voltage in kV
        :param defocus_u:   Defocus depth along the u-axis in angstrom
        :param defocus_v:   Defocus depth along the v-axis in angstrom
        :param defocus_ang: Angle between the x-axis and the u-axis in radians
        :param Cs:          Spherical aberration constant
        :param alpha:       Amplitude contrast phase in radians
        :param B:           Envelope decay in inverse square angstrom (default 0)
        """
        super().__init__(dim=2, radial=defocus_u == defocus_v)
        self.pixel_size = pixel_size
        self.voltage = voltage
        self.wavelength = voltage_to_wavelength(self.voltage)
        self.defocus_u = defocus_u
        self.defocus_v = defocus_v
        self.defocus_ang = defocus_ang
        self.Cs = Cs
        self.alpha = alpha
        self.B = B

        self.defocus_mean = 0.5 * (self.defocus_u + self.defocus_v)
        self.defocus_diff = 0.5 * (self.defocus_u - self.defocus_v)

    def _evaluate(self, omega):
        om_y, om_x = np.vsplit(omega / (2 * np.pi * self.pixel_size), 2)

        eps = np.finfo(np.pi).eps
        ind_nz = (np.abs(om_x) > eps) | (np.abs(om_y) > eps)
        angles_nz = np.arctan2(om_y[ind_nz], om_x[ind_nz])
        angles_nz -= self.defocus_ang

        defocus = np.zeros_like(om_x)
        defocus[ind_nz] = self.defocus_mean + self.defocus_diff * np.cos(2 * angles_nz)

        c2 = -np.pi * self.wavelength * defocus
        c4 = 0.5 * np.pi * (self.Cs * 1e7) * self.wavelength**3

        r2 = om_x**2 + om_y**2
        r4 = r2**2
        gamma = c2 * r2 + c4 * r4
        h = np.sqrt(1 - self.alpha**2) * np.sin(gamma) - self.alpha * np.cos(gamma)

        # For historical reference, below is a translated formula from the legacy MATLAB code.
        # The two implementations seem to agree for odd images, but the original MATLAB code
        # behaves differently for even image sizes.
        # h = np.sin(c2*r2 + c4*r2*r2 - self.alpha)

        if self.B:
            h *= np.exp(-self.B * r2)

        return h.squeeze()

    def scale(self, c=1):
        return CTFFilter(
            pixel_size=self.pixel_size * c,
            voltage=self.voltage,
            defocus_u=self.defocus_u,
            defocus_v=self.defocus_v,
            defocus_ang=self.defocus_ang,
            Cs=self.Cs,
            alpha=self.alpha,
            B=self.B,
        )


class RadialCTFFilter(CTFFilter):
    def __init__(
        self, pixel_size=10, voltage=200, defocus=15000, Cs=2.26, alpha=0.07, B=0
    ):
        super().__init__(
            pixel_size=pixel_size,
            voltage=voltage,
            defocus_u=defocus,
            defocus_v=defocus,
            defocus_ang=0,
            Cs=Cs,
            alpha=alpha,
            B=B,
        )



import logging
from abc import ABC, abstractmethod

import numpy as np

logger = logging.getLogger(__name__)


class Class2D(ABC):
    """
    Base class for 2D Image Classification methods.
    """

    def __init__(
        self,
        src,
        n_nbor=100,
        seed=None,
        dtype=None,
    ):
        """
        Base constructor of an object for classifying 2D images.

        :param src: ImageSource or subclass, provides images.
        :param n_nbor: Number of nearest neighbors to compute.
        :param seed: Optional RNG seed to be passed to random methods, (example Random NN).
        :param dtype: Numpy dtype, defaults to `src.dtype`.
        """
        self.src = src

        if dtype is not None:
            self.dtype = np.dtype(dtype)
            if self.dtype != self.src.dtype:
                logger.warning(
                    f"Class2D src.dtype {self.src.dtype} does not match self.dtype {self.dtype}."
                )
        else:
            self.dtype = self.src.dtype

        self.n_nbor = n_nbor
        self.seed = seed

    @abstractmethod
    def classify(self):
        """
        Classify the images from Source into classes with similar viewing angles.

        Returns classes and associated metadata (classes, reflections, distances)
        """



import pkgutil

from click.core import Command, Group

import aspire.commands


def main_entry():
    main = Group(chain=False)

    # TODO: Add options
    # @click.option('--debug/--no-debug', default=False, help="Default is --no-debug.")
    # @click.option('-v', '--verbosity', default=0, help='Verbosity level (0-3).')

    for importer, modname, _ in pkgutil.iter_modules(aspire.commands.__path__):
        module = importer.find_module(modname).load_module(modname)
        commands = [v for v in module.__dict__.values() if isinstance(v, Command)]
        for command in commands:
            main.add_command(command)

    main.main(prog_name="aspire")


if __name__ == "__main__":
    main_entry()



import logging

from aspire.abinitio import CLOrient3D

logger = logging.getLogger(__name__)


class CommLineGCAR(CLOrient3D):
    """
    Define a derived class to estimate 3D orientations using Globally Consistent Angular Reconstitution described
    as below:
    R. Coifman, Y. Shkolnisky, F. J. Sigworth, and A. Singer, Reference Free Structure Determination through
    Eigenvestors of Center of Mass Operators, Applied and Computational Harmonic Analysis, 28, 296-312 (2010).

    """

    def __init__(self, src):
        """
        constructor of an object for estimating 3D oreintations
        """
        pass

    def estimate(self):
        """
        perform estimation of orientations
        """
        pass

    def output(self):
        """
        Output the 3D orientations
        """
        pass



import logging

import matplotlib.pyplot as plt
import numpy as np
from sklearn.neighbors import NearestNeighbors

from aspire.basis import FSPCABasis
from aspire.classification import Class2D
from aspire.classification.legacy_implementations import bispec_2drot_large, pca_y
from aspire.numeric import ComplexPCA
from aspire.utils import trange
from aspire.utils.random import rand

logger = logging.getLogger(__name__)


class RIRClass2D(Class2D):
    def __init__(
        self,
        src,
        pca_basis=None,
        fspca_components=None,
        alpha=1 / 3,
        sample_n=4000,  # Paper had 4000, but MATLAB code suggested 50000
        bispectrum_components=300,
        n_nbor=100,
        bispectrum_freq_cutoff=None,
        large_pca_implementation="legacy",
        nn_implementation="legacy",
        bispectrum_implementation="legacy",
        batch_size=512,
        dtype=None,
        seed=None,
    ):
        """
        Constructor of an object for classifying 2D images using
        Rotationally Invariant Representation (RIR) algorithm.

        At a high level this consumes a Source instance `src`,
        and a FSPCA Basis `pca_basis`.

        Yield class averages by first performing `classify`,
        then performing `output`.

        Z. Zhao, Y. Shkolnisky, A. Singer, Rotationally Invariant Image Representation
        for Viewing Direction Classification in Cryo-EM. (2014)

        :param src: Source instance, for classification.
        :param pca_basis: Optional FSPCA Basis instance
        :param fspca_components: Optinally set number of components (top eigvals) to keep from full FSCPA.
            Default value of None will infer from `pca_basis` when provided, otherwise defaults to 400.
        :param alpha: Amplitude Power Scale, default 1/3 (eq 20 from  RIIR paper).
        :param sample_n: Threshold for random sampling of bispectrum coefs. Default 4000,
            high values such as 50000 reduce random sampling.
        :param n_nbor: Number of nearest neighbors to compute.
        :param bispectrum_freq_cutoff: Truncate (zero) high k frequecies above (int) value, defaults off (None).
        :param large_pca_implementation: See `pca`.
        :param nn_implementation: See `nn_classification`.
        :param bispectrum_implementation: See `bispectrum`.
        :param batch_size: Chunk size (typically number of images) for batched methods.
        :param dtype: Optional dtype, otherwise taken from src.
        :param seed: Optional RNG seed to be passed to random methods, (example Random NN).
        :return: RIRClass2D instance to be used to compute bispectrum-like rotationally invariant 2D classification.
        """

        super().__init__(
            src=src,
            n_nbor=n_nbor,
            seed=seed,
            dtype=dtype,
        )
        self.batch_size = int(batch_size)

        # Implementation Checks
        # # Do we have a sane Nearest Neighbor
        nn_implementations = {
            "legacy": self._legacy_nn_classification,
            "sklearn": self._sk_nn_classification,
        }
        if nn_implementation not in nn_implementations:
            raise ValueError(
                f"Provided nn_implementation={nn_implementation} not in {nn_implementations.keys()}"
            )
        self._nn_implementation = nn_implementation  # Save str for logger
        self._nn_classification = nn_implementations[nn_implementation]

        # # Do we have a sane Large Dataset PCA
        large_pca_implementations = {
            "legacy": self._legacy_pca,
            "sklearn": self._sk_pca,
        }
        if large_pca_implementation not in large_pca_implementations:
            raise ValueError(
                f"Provided large_pca_implementation={large_pca_implementation} not in {large_pca_implementations.keys()}"
            )
        self._pca = large_pca_implementations[large_pca_implementation]

        # # Do we have a sane Bispectrum component
        bispectrum_implementations = {
            "legacy": self._legacy_bispectrum,
            "devel": self._devel_bispectrum,
        }
        if bispectrum_implementation not in bispectrum_implementations:
            raise ValueError(
                f"Provided bispectrum_implementation={bispectrum_implementation} not in {bispectrum_implementations.keys()}"
            )
        elif bispectrum_implementation == "legacy" and self._pca != self._legacy_pca:
            raise ValueError(
                '"legacy" bispectrum_implementation implies'
                ' large_pca_implementation="legacy".'
                " Check class configuration and retry."
            )
        self._bispectrum = bispectrum_implementations[bispectrum_implementation]

        # For now, only run with FSPCA basis
        if pca_basis and not isinstance(pca_basis, FSPCABasis):
            raise NotImplementedError(
                "RIRClass2D has currently only been developed for pca_basis as a FSPCABasis."
            )
        self.pca_basis = pca_basis

        # When a user provides a `pca_basis` and `fspca_components`
        #  the arg is either redundant (same) or conflicting with `pca_basis`.
        if pca_basis and fspca_components is not None:
            # Check the provided components match. On match we can use this value.
            if pca_basis.components != fspca_components:
                raise RuntimeError(
                    f"`pca_basis` components {pca_basis.components} != {fspca_components} `fspca_components` provided by user."
                )
        elif pca_basis:  # infer `fspca_components` from pca_basis components
            fspca_components = pca_basis.components

            # For small problems, such as unit tests, we also need to guard against
            #   requesting more fspca_components than exist in the basis now,
            # In the case RIRClass2d instantiates the pca_basis,
            #   this will be checked then, in FSPCABasis.
            pca_basis._check_components()
        elif fspca_components is None:
            # Default of 400 components was taken from legacy reearch and code.
            fspca_components = 400

        self.fspca_components = fspca_components
        self.bispectrum_components = bispectrum_components
        # Similarly, for small problems we need to check these counts.
        if fspca_components < bispectrum_components:
            raise RuntimeError(
                f"fspca_components {fspca_components} < bispectrum components {bispectrum_components}."
                "  Reduce bispectrum_components. Reasonable starting value is int(0.75*fspca_components)."
            )

        self.sample_n = sample_n
        self.alpha = alpha
        self.bispectrum_freq_cutoff = bispectrum_freq_cutoff

        if self.src.n < self.bispectrum_components:
            raise RuntimeError(
                f"{self.src.n} Images too small for Bispectrum Components {self.bispectrum_components}."
                "  Increase number of images or reduce components."
            )

    def classify(self, diagnostics=False):
        """
        This is the high level method to perform the 2D images classification.

        The stages of this method are intentionally modular so they may be
        swapped for other implementations.

        :param diagnostics: Optionally plots distribution of distances
        """

        # # Stage 1: Compute coef and reduce dimensionality.
        if self.pca_basis is None:
            self.pca_basis = FSPCABasis(
                self.src, components=self.fspca_components, batch_size=self.batch_size
            )

        # For convenience, assign the fb_basis used in the pca_basis.
        self.fb_basis = self.pca_basis.basis

        # Get the expanded coefs in the compressed FSPCA space.
        self.fspca_coef = self.pca_basis.spca_coef

        # Compute Bispectrum
        coef_b, coef_b_r = self.bispectrum(self.fspca_coef)

        # # Stage 2: Compute Nearest Neighbors
        logger.info(f"Calculate Nearest Neighbors using {self._nn_implementation}.")
        self.classes, self.reflections, self.distances = self.nn_classification(
            coef_b, coef_b_r
        )

        if diagnostics:
            # Lets peek at the distribution of distances
            # zero index is self, distance 0, ignored
            plt.hist(self.distances[:, 1:].flatten(), bins="auto")
            plt.show()

            # Report some information about reflections
            logger.info(
                f"Count reflected: {np.sum(self.reflections)}"
                f" {100 * np.mean(self.reflections) } %"
            )

        return self.classes, self.reflections, self.distances

    def pca(self, M):
        """
        Any PCA implementation here should return both
        coef_b and coef_b_r that are (n_img, n_components).

        `n_components` is typically self.bispectrum_components.
        However, for small problems it may return `n_components`=`n_img`,
        since that would be the smallest dimension.

        To extend class with an additional PCA like method,
        add as private method and list in `large_pca_implementations`.

        :param M: Array (n_img, m_features), typically complex.
        :returns: Tuple of arrays coef_b coef_b_r.
        """
        # _pca is assigned during initialization.
        return self._pca(M)

    def nn_classification(self, coef_b, coef_b_r):
        """
        Takes in features as pair of arrays (coef_b coef_b_r),
        each having shape (n_img, features)
        where features = min(self.bispectrum_components, n_img).

        Result is array (n_img, n_nbor) with entry `i` representing
        index `i` into class input img array (src).

        To extend with an additonal Nearest Neighbor algo,
        add as a private method and list in nn_implementations.

        :param coef_b:
        :param coef_b_r:
        :returns:  Tuple of classes, refl, dists where
            classes is an integer array of indices representing image ids,
            refl is a bool array representing reflections (True is refl),
            and distances is an array of distances as returned by NN implementation.
        """
        # _nn_classification is assigned during initialization.
        return self._nn_classification(coef_b, coef_b_r)

    def bispectrum(self, coef):
        """
        All bispectrum implementations should consume a stack of fspca coef
        and return bispectrum coefficients.

        :param coef: complex steerable coefficients (eg. from FSPCABasis).
        :returns: tuple of arrays (coef_b, coef_b_r)
        """
        # _bispectrum is assigned during initialization.
        return self._bispectrum(coef)

    def _sk_nn_classification(self, coeff_b, coeff_b_r):
        """
        Perform nearest neighbor classification using scikit learn.

        Note "distances" are as computed by scikit, defaults to Euclidean.
        """

        n_img = self.src.n

        # Third party tools generally expecting:
        #   slow axis as n_data, fast axis n_features.
        # Also most third party NN complain about complex...
        #   so we'll pretend we have 2*n_features of real values.
        # Don't worry about the copy because NearestNeighbors wants
        #   C-contiguous anyway... (it would copy internally otherwise)
        X = np.column_stack((coeff_b.real, coeff_b.imag))
        # We'll also want to consider the neighbors under reflection.
        #   These coefficients should be provided by coeff_b_r
        X_r = np.column_stack((coeff_b_r.real, coeff_b_r.imag))

        # We can compare both non-reflected and reflected representations as one large set by
        #   taking care later that we store refl=True where indices>=n_img
        X_both = np.concatenate((X, X_r))

        nbrs = NearestNeighbors(n_neighbors=self.n_nbor, algorithm="auto").fit(X_both)
        distances, indices = nbrs.kneighbors(X)

        # There were two sets of vectors each n_img long.
        #   The second set represented reflected.
        #   When a reflected coef vector is a nearest neighbor,
        #   we notate the original image index (indices modulus n_img),
        #   and notate we'll need the reflection (refl).
        classes = indices % n_img
        refl = np.array(indices // n_img, dtype=bool)

        return classes, refl, distances

    def _legacy_nn_classification(self, coeff_b, coeff_b_r):
        """
        Perform nearest neighbor classification using port of ASPIRE legacy MATLAB code.

        Note `distances` returned from this method are dot products, ie "corr".
        """

        # Note kept ordering from legacy code (n_features, n_img)
        coeff_b = coeff_b.T
        coeff_b_r = coeff_b_r.T

        n_im = self.src.n
        # Shouldn't have more neighbors than images
        n_nbor = self.n_nbor
        if n_nbor >= n_im:
            logger.warning(
                f"Requested {self.n_nbor} self.n_nbor, but only {n_im} images. Setting self.n_nbor={n_im-1}."
            )
            n_nbor = n_im - 1

        concat_coeff = np.concatenate((coeff_b, coeff_b_r), axis=1)

        num_batches = (n_im + self.batch_size - 1) // self.batch_size

        classes = np.zeros((n_im, n_nbor), dtype=int)
        distances = np.zeros((n_im, n_nbor), dtype=self.dtype)

        for i in trange(num_batches):
            start = i * self.batch_size
            finish = min((i + 1) * self.batch_size, n_im)
            batch = np.conjugate(coeff_b[:, start:finish])
            corr = np.real(np.dot(batch.T, concat_coeff))

            assert np.all(
                np.abs(corr) <= 1.01  # Allow some numerical wiggle
            ), f"Corr out of [-1,1] bounds {np.min(corr)} {np.max(corr)}."
            # Clamp
            corr = np.maximum(corr, -1)
            corr = np.minimum(corr, 1)

            # Note legacy did not include the original image?
            # classes[start:finish] = np.argsort(-corr, axis=1)[:, 1 : n_nbor + 1]
            # This now does include the original image
            # (Matches sklean implementation.)
            #
            # Also we've converted from correlation to distance=1-correlation
            # https://github.com/ComputationalCryoEM/ASPIRE-Python/discussions/867
            dist = np.sqrt(2.0 - 2.0 * corr)

            classes[start:finish] = np.argsort(dist, axis=1)[:, :n_nbor]
            # Store the corr values for the n_nbors in this batch
            distances[start:finish] = np.take_along_axis(
                dist, classes[start:finish], axis=1
            )

        # There were two sets of vectors each n_img long.
        #   The second set represented reflected.
        #   When a reflected coef vector is a nearest neighbor,
        #   we notate the original image index (indices modulus n_img),
        #   and notate we'll need the reflection (refl).
        refl = np.array(classes // n_im, dtype=bool)
        classes %= n_im

        return classes, refl, distances

    def _legacy_pca(self, M):
        """
        This is more or less the historic implementation ported
        to Python from code calling MATLAB's `pca_y`.
        """

        # ### The following was from legacy code. Be careful wrt order.
        M = M.T
        u, s, v = pca_y(M, self.bispectrum_components, seed=self.seed)

        # Contruct coefficients
        coef_b = np.einsum("i, ij -> ij", s, np.conjugate(v))
        coef_b_r = np.conjugate(u.T).dot(np.conjugate(M))

        # Normalize
        coef_b /= np.linalg.norm(coef_b, axis=0)
        coef_b_r /= np.linalg.norm(coef_b_r, axis=0)

        # Transpose (this code was originally F order)
        coef_b = coef_b.T
        coef_b_r = coef_b_r.T

        return coef_b, coef_b_r

    def _sk_pca(self, M):
        # Avoiding SK directly for now,
        #   while it is really useful, it
        #   expects real data.
        #   We use an extension of SK that is hacked to admit complex.
        pca = ComplexPCA(
            self.bispectrum_components,
            copy=False,  # careful, overwrites data matrix... we'll handle the copies.
            svd_solver="auto",  # use randomized (Halko) for larger problems
            random_state=self.seed,
        )
        coef_b = pca.fit_transform(M.copy())
        coef_b_r = coef_b.conj()

        # I'm also not sure why this norm is needed...
        #  but it does work better with it.
        coef_b /= np.linalg.norm(coef_b, axis=1)[:, np.newaxis]
        coef_b_r /= np.linalg.norm(coef_b_r, axis=1)[:, np.newaxis]

        return coef_b, coef_b_r

    def _devel_bispectrum(self, coef):
        coef = self.pca_basis.to_complex(coef)
        # Take just positive frequencies, corresponds to complex indices.
        # Original implementation used norm of Complex values, here abs of Real.
        eigvals = np.abs(self.pca_basis.eigvals[self.pca_basis.signs_indices >= 0])

        # Legacy code included a sanity check:
        # non_zero_freqs = self.pca_basis.complex_angular_indices != 0
        # coef_norm = np.log(np.power(np.abs(coef[:,non_zero_freqs]), self.alpha)).all())
        # just stick to the paper (eq 20) for now , look at this more later.

        coef_normed = np.where(
            coef == 0, 0, coef / np.power(np.abs(coef), 1 - self.alpha)
        )  # should use an epsilon here...

        if not np.isfinite(coef_normed).all():
            raise ValueError("Coefs should be finite")

        # ## Compute and reduce Bispectrum
        m = np.power(eigvals, self.alpha)
        m = m[
            self.pca_basis.complex_angular_indices != 0
        ]  # filter non_zero_freqs eq 18,19
        pm = m / np.sum(m)
        x = rand(len(m))
        m_mask = x < self.sample_n * pm

        M = None

        for i in trange(self.src.n):
            B = self.pca_basis.calculate_bispectrum(
                coef_normed[i, np.newaxis],
                filter_nonzero_freqs=True,
                freq_cutoff=self.bispectrum_freq_cutoff,
            )

            # ### Truncate Bispectrum (by sampling)
            # ### Note, where is this written down? (and is it even needed?)
            # It is only briefly mentioned in the paper and was more/less
            # soft disabled in the matlab code.
            B = B[m_mask][:, m_mask]
            logger.info(f"Truncating Bispectrum to {B.shape} ({np.size(B)}) coefs.")

            # B is symmetric, take lower triangle of first two axis.
            tril = np.tri(B.shape[0], dtype=bool)
            B = B[tril, :]
            logger.info(f"Symmetry reduced Bispectrum to {np.size(B)} coefs.")
            # B is sparse and should have same sparsity for any image up to underflows...
            B = B.ravel()[np.flatnonzero(B)]
            logger.info(f"Sparse (nnz) reduced Bispectrum to {np.size(B)} coefs.")

            # Legacy code had bispect flattened as CSR and some other hacks.
            #   For now, we'll compute it densely then take nonzeros.
            if M is None:
                # Instanstiate M with B's nnz size
                M = np.empty((self.src.n, B.shape[0]), dtype=coef.dtype)
            M[i] = B

        # Reduce dimensionality of Bispectrum sample with PCA
        logger.info(
            f"Computing Large PCA, returning {self.bispectrum_components} components."
        )
        # should add memory sanity check here... these can be crushingly large...
        coef_b, coef_b_r = self.pca(M)

        return coef_b, coef_b_r

    def _legacy_bispectrum(self, coef, retry_attempts=3):
        """
        This code was ported to Python by an unkown author,
        and is the closest viable reference material.

        :param coef: Real valued basis coefficients.
        :param retry_attempts: Optional, max attempts to retry randomized
            bispec_2drot_large.  Defaults to 3.

        :return: Compressed feature and reflected feature vectors.
        """

        # The legacy code expects the complex representation
        coef = self.pca_basis.to_complex(coef)
        complex_eigvals = self.pca_basis.to_complex(self.pca_basis.eigvals).reshape(
            self.pca_basis.complex_count
        )  # flatten

        # bispec_2drot_large has a random selection component.
        # Sometimes this can fail to return a complete feature vector.
        # In this case we can retry, but if not successful raise an error.
        # This seems to occur more frequently at very low resolutions (<=32),
        # and likely requires tuning other RIR parameters for small problems.
        attempt = 0
        while attempt < retry_attempts:
            attempt += 1
            coef_b, coef_b_r = bispec_2drot_large(
                coeff=coef.T,  # Note F style transpose here and in return
                freqs=self.pca_basis.complex_angular_indices,
                eigval=complex_eigvals,
                alpha=self.alpha,
                sample_n=self.sample_n,
                seed=self.seed,
            )
            # If we have produced a feature vector
            if coef_b.size != 0:
                break  # Return feature vector.

        # while-else: we've exceeded retry attempts.
        else:
            raise RuntimeError(
                "bispec_2drot_large failed to return valid feature vector."
                f" Returned {coef_b.shape} after {attempt} attempts."
            )

        return coef_b.T, coef_b_r.T



import logging

import numpy as np

from aspire import config
from aspire.utils import LogFilterByCount, complex_type, real_type

logger = logging.getLogger(__name__)

# Cached Plan Class objects, indexed by backend string identifier, and ordered by preference (highest first)
# The values are either Plan subclasses (for working backends), or None (for non-working backends)
# Populated by 'check_backends()' when first needed.
backends = None
# Default preferred Plan subclass
default_plan_class = None


def check_backends(raise_errors=True):
    """
    Check all NFFT backends in package configuration
    :param raise_errors: Whether to raise a RuntimeError if no backends detected.

    :return: On return, the global names 'backends'/'default_plan_class' have been populated
    """

    global backends, default_plan_class

    def _try_backend(backend):
        """
        This function tries out a particular NFFT backend by name.
        :param backend: A string representing the NFFT backend we want to try. Currently one of:
            'cufinufft'
                The Python wrapper for the CUDA variant of FINUFFT library
                https://github.com/flatironinstitute/cufinufft
            'finufft'
                The Python wrapper for the FlatIron Institute's FINUFFT library
                https://github.com/flatironinstitute/finufft
            'pynfft'
                The Python wrapper for the Chemnitz NFFT library
                https://www-user.tu-chemnitz.de/~potts/nfft/
        :return: The proper Plan-subclass if the backend is expected to work or None otherwise.

        It's important to keep these checks lightweight since all usable backend classes are cached on module load.
        We only check for working imports here to keep things simple and lightweight.
        If imports are working but the actual backend is not then we have bigger problems anyway.
        """

        logger.debug(f"Trying NFFT backend {backend}")
        plan_class = None
        msg = None
        if backend == "cufinufft":
            try:
                # Note we expect the cu library is in LD path.
                #   That is managed by pip if using a wheel,
                #   or the user if they have built from source.
                from aspire.nufft.cufinufft import CufinufftPlan

                plan_class = CufinufftPlan
            except Exception as e:
                msg = str(e)

        elif backend == "finufft":
            try:
                from finufft import Plan  # noqa: F401

                from aspire.nufft.finufft import FinufftPlan

                plan_class = FinufftPlan
            except Exception as e:
                msg = str(e)

        elif backend == "pynfft":
            try:
                from pynfft.nfft import NFFT  # noqa: F401

                from aspire.nufft.pynfft import PyNfftPlan

                plan_class = PyNfftPlan
            except Exception as e:
                msg = str(e)

        if plan_class is None:
            logger.debug(f"NFFT backend {backend} not usable:\n\t{msg}")
        else:
            logger.debug(f"NFFT backend {backend} usable.")
            return plan_class

    # Note this dictionary is intentionally ordered
    backends = {k: _try_backend(k) for k in config["nufft"]["backends"].as_str_seq()}
    try:
        default_backend = next(k for k, v in backends.items() if v is not None)
        logger.debug(f"Selected NFFT backend = {default_backend}.")
        default_plan_class = backends[default_backend]
    except StopIteration:
        msg = "No usable NFFT backend detected."
        logger.error(msg)
        default_plan_class = None
        if raise_errors:
            raise RuntimeError(msg) from None


def all_backends():
    """
    Determine all available NFFT backends

    :return: A list of strings representing available NFFT backends
    """
    global backends

    if backends is None:
        check_backends(raise_errors=False)
    return [k for k, v in backends.items() if v is not None]


def backend_available(backend):
    """
    Whether a particular NFFT backend is available

    :param backend: String representing the NFFT backend, e.g. 'finufft' or 'pynfft'
    :return: Boolean on whether the backend is available
    """
    return backend in all_backends()


class Plan:
    # TODO: move common functionality up the hierarchy
    def __new__(cls, *args, **kwargs):
        if cls is Plan:
            if "backend" in kwargs:
                if backend_available(kwargs["backend"]):
                    return super(Plan, cls).__new__(backends[kwargs["backend"]])
                else:
                    raise RuntimeError("Requested backend unavailable")
            else:
                # If a Plan was constructed as a generic Plan(), use the default (best) Plan class
                if default_plan_class is None:
                    # Limit log noise to once
                    with LogFilterByCount(logger, 1):
                        check_backends(raise_errors=True)
                return super(Plan, cls).__new__(default_plan_class)
        else:
            # If a Plan-subclass was constructed directly, invoke default behavior
            return super(Plan, cls).__new__(cls)


def anufft(sig_f, fourier_pts, sz, real=False, epsilon=1e-8):
    """
    Wrapper for 1, 2, and 3 dimensional Non Uniform FFT Adjoint.
    Dimension is based on the dimension of fourier_pts and checked against sig_f.

    Selects best available package from `nfft` `backends` configuration list.

    :param sig_f: Array representing the signal(s) in Fourier space to be transformed. \
    sig_f either matches length of fourier_pts or sig_f.shape is stack of (`ntransforms`, ...).
    :param fourier_pts: The points in Fourier space where the Fourier transform is to be calculated,
            arranged as a dimension-by-K array. These need to be in the range [-pi, pi] in each dimension.
    :param sz: A tuple indicating the geometry of the signal.
    :param real: Optional Bool indicating if you would like only the real components, Defaults False.
    :return: The Non Uniform FFT adjoint transform.

    """

    if fourier_pts.dtype != real_type(sig_f.dtype):
        raise RuntimeError(
            "anufft passed inconsistent dtypes."
            f" fourier_pts: {fourier_pts.dtype}"
            f" does not match precision of signal data: {sig_f.dtype}."
        )

    if sig_f.dtype != complex_type(sig_f.dtype):
        logger.debug("anufft passed real_type for signal, converting")
        sig_f = sig_f.astype(complex_type(sig_f.dtype))

    ntransforms = 1
    if len(sig_f.shape) == 2:
        ntransforms = sig_f.shape[0]

    plan = Plan(
        sz=sz, fourier_pts=fourier_pts, ntransforms=ntransforms, epsilon=epsilon
    )
    adjoint = plan.adjoint(sig_f)
    return np.real(adjoint) if real else adjoint


def nufft(sig_f, fourier_pts, real=False, epsilon=1e-8):
    """
    Wrapper for 1, 2, and 3 dimensional Non Uniform FFT
    Dimension is based on the dimension of fourier_pts and checked against sig_f.

    Selects best available package from `nfft` `backends` configuration list.

    :param sig_f: Array representing the signal(s) in real space to be transformed. \
    sig_f either matches `sz` or sig_f.shape is stack of (..., `ntransforms`).
    :param fourier_pts: The points in Fourier space where the Fourier transform is to be calculated,
            arranged as a dimension-by-K array. These need to be in the range [-pi, pi] in each dimension.
    :param real: Optional Bool indicating if you would like only the real components, Defaults False.
    :return: The Non Uniform FFT transform.

    """

    if fourier_pts.dtype != real_type(sig_f.dtype):
        raise RuntimeError(
            "nufft passed inconsistent dtypes."
            f" fourier_pts: {fourier_pts.dtype}"
            f" does not match precision of signal data: {sig_f.dtype}."
        )

    if sig_f.dtype != complex_type(sig_f.dtype):
        logger.debug("nufft passed real_type for signal, converting")
        sig_f = sig_f.astype(complex_type(sig_f.dtype))

    # Unpack the dimension of the signal
    #   Note, infer dimension from fourier_pts because signal might be a stack.
    dimension = fourier_pts.shape[0]

    # Unpack the resolution of the signal
    resolution = sig_f.shape[1]

    # Construct tuple describing geometry of signal
    sz = (resolution,) * dimension

    ntransforms = 1
    if len(sig_f.shape) == dimension + 1:
        ntransforms = sig_f.shape[0]

    plan = Plan(
        sz=sz, fourier_pts=fourier_pts, ntransforms=ntransforms, epsilon=epsilon
    )
    transform = plan.transform(sig_f)
    return np.real(transform) if real else transform



import cupy as cp

from aspire.numeric.base_fft import FFT


class CupyFFT(FFT):
    """
    Define a unified wrapper class for Cupy FFT functions

    To be consistent with Scipy and Pyfftw, not all arguments are included.
    """

    def fft(self, x, axis=-1, workers=-1):
        return cp.fft.fft(x, axis=axis)

    def ifft(self, x, axis=-1, workers=-1):
        return cp.fft.ifft(x, axis=axis)

    def fft2(self, x, axes=(-2, -1), workers=-1):
        return cp.fft.fft2(x, axes=axes)

    def ifft2(self, x, axes=(-2, -1), workers=-1):
        return cp.fft.ifft2(x, axes=axes)

    def fftn(self, x, axes=None, workers=-1):
        return cp.fft.fftn(x, axes=axes)

    def ifftn(self, x, axes=None, workers=-1):
        return cp.fft.ifftn(x, axes=axes)

    def fftshift(self, x, axes=None):
        return cp.fft.fftshift(x, axes=axes)

    def ifftshift(self, x, axes=None):
        return cp.fft.ifftshift(x, axes=axes)



import numpy as np
from pynfft.nfft import NFFT

from aspire.nufft import Plan
from aspire.nufft.utils import nextpow2


class PyNfftPlan(Plan):
    @staticmethod
    def epsilon_to_nfft_cutoff(epsilon):
        # NOTE: These are obtained empirically. Should have a theoretical derivation.
        rel_errs = [6e-2, 2e-3, 2e-5, 2e-7, 3e-9, 4e-11, 4e-13, 0]
        return list(
            filter(lambda i_err: i_err[1] < epsilon, enumerate(rel_errs, start=1))
        )[0][0]

    def __init__(self, sz, fourier_pts, epsilon=1e-15, **kwargs):
        """
        A plan for non-uniform FFT (3D)
        :param sz: A tuple indicating the geometry of the signal
        :param fourier_pts: The points in Fourier space where the Fourier transform is to be calculated,
            arranged as a 3-by-K array. These need to be in the range [-pi, pi] in each dimension.
        :param epsilon: The desired precision of the NUFFT
        """
        self.sz = sz
        self.dim = len(sz)
        self.fourier_pts = fourier_pts
        self.num_pts = fourier_pts.shape[1]
        self.epsilon = epsilon

        self.cutoff = PyNfftPlan.epsilon_to_nfft_cutoff(epsilon)
        self.multi_bandwith = tuple(2 * 2 ** nextpow2(self.sz))
        # TODO - no other flags used in the MATLAB code other than these 2 are supported by the PyNFFT wrapper
        self._flags = ("PRE_PHI_HUT", "PRE_PSI")

        self._plan = NFFT(
            N=self.sz,
            M=self.num_pts,
            n=self.multi_bandwith,
            m=self.cutoff,
            flags=self._flags,
        )

        self._plan.x = ((1.0 / (2 * np.pi)) * self.fourier_pts).T
        self._plan.precompute()

    def transform(self, signal):
        assert (
            signal.shape == self.sz
        ), f"Signal to be transformed must have shape {self.sz}"

        self._plan.f_hat = signal.astype("complex64")
        f = self._plan.trafo()

        if signal.dtype == np.float32:
            f = f.astype("complex64")

        return f

    def adjoint(self, signal):
        self._plan.f = signal.astype("complex64")
        f_hat = self._plan.adjoint()

        if signal.dtype == np.float32:
            f_hat = f_hat.astype("complex64")

        return f_hat



import numpy as np
import scipy.sparse as sparse


def transform_complex_to_real(B_conj, ells):
    """
    Transforms coefficients of the matrix B (see Eq. 3) from complex
        to real. B is the linear transformation that takes FLE coefficients
        to images.

    :param B_conj: Complex conjugate of the matrix B.
    :param ells: List of ells (Bessel function orders) in this basis.
    :return: Transformed matrix.
    """
    num_basis_functions = B_conj.shape[1]
    X = np.zeros(B_conj.shape, dtype=np.float64)

    for i in range(num_basis_functions):
        ell = ells[i]
        if ell == 0:
            X[:, i] = np.real(B_conj[:, i])
        # for each ell != 0, we can populate two entries of the matrix
        # by taking the complex conjugate of the ell with the opposite sign
        if ell < 0:
            s = (-1) ** np.abs(ell)
            x0 = (B_conj[:, i] + s * B_conj[:, i + 1]) / np.sqrt(2)
            x1 = (-B_conj[:, i] + s * B_conj[:, i + 1]) / (1j * np.sqrt(2))
            X[:, i] = np.real(x0)
            X[:, i + 1] = np.real(x1)

    return X


def precomp_transform_complex_to_real(ells):
    """
    Returns a sparse matrix that transforms coefficients into the complex
        representation of the basis to coefficients in the real
        representation of the basis. See Remark 1.1 of Marshall, Mickelin,
        and Singer.

    :param ells: The list of integer Bessel function orders.
    :return: Sparse complex to real transformation matrix.
    """
    count = len(ells)
    num_nonzero = np.sum(ells == 0) + 2 * np.sum(ells != 0)
    idx = np.zeros(num_nonzero, dtype=int)
    jdx = np.zeros(num_nonzero, dtype=int)
    vals = np.zeros(num_nonzero, dtype=np.complex128)

    k = 0
    for i in range(count):
        ell = ells[i]
        # ell = 0 is a special case (DC component)
        if ell == 0:
            vals[k] = 1
            idx[k] = i
            jdx[k] = i
            k = k + 1
        # Only branch the case ell < 0 and also update -ell
        # via complex conjugation
        if ell < 0:
            s = (-1) ** np.abs(ell)

            # positive ell
            vals[k] = 1 / np.sqrt(2)
            idx[k] = i
            jdx[k] = i
            k = k + 1

            # positive ell
            vals[k] = s / np.sqrt(2)
            idx[k] = i
            jdx[k] = i + 1
            k = k + 1

            # negative ell
            vals[k] = -1 / (1j * np.sqrt(2))
            idx[k] = i + 1
            jdx[k] = i
            k = k + 1

            # negative ell
            vals[k] = s / (1j * np.sqrt(2))
            idx[k] = i + 1
            jdx[k] = i + 1
            k = k + 1

    A = sparse.csr_matrix((vals, (idx, jdx)), shape=(count, count), dtype=np.complex128)

    return A


def barycentric_interp_sparse(target_points, known_points, numsparse):
    """
    Returns the sparse matrices that perform barycentric interpolation to compute values
        of Betas at the points `target_points` at known points `known_points`, and the transpose
        of this operation. For each target point in `target_points`, only `numsparse` centered
        source points from `known_points` around the target point are used.

        Performed via the method described in

        "Barycentric Lagrange Interpolation", Jean-Paul Berrut and Lloyd Trefethen.
        SIAM Review 2004 46:3, 501-517
        https://people.maths.ox.ac.uk/trefethen/barycentric.pdf

    :param target_points: The target set of points at which to evaluate the functions.
    :param known_points: The points at which the values of the functions are known.
    :param numsparse: Number of points used for interpolation around each target point.
    :return: The interpolation matrix and its transpose as a 2-tuple.
    """

    n = len(target_points)
    m = len(known_points)

    # Modify points by 2e-16 to avoid division by zero
    vals, x_ind, xs_ind = np.intersect1d(
        target_points, known_points, return_indices=True, assume_unique=True
    )
    target_points[x_ind] = target_points[x_ind] + 2e-16

    idx = np.zeros((n, numsparse))
    jdx = np.zeros((n, numsparse))
    vals = np.zeros((n, numsparse))
    xss = np.zeros((n, numsparse))
    denom = np.zeros((n, 1))
    temp = np.zeros((n, 1))
    ws = np.zeros((n, numsparse))
    xdiff = np.zeros(n)

    # loop over target points
    for i in range(n):
        # choose `numsparse` source points centered around each target point
        # in order to apply a sparse barycentric interpolation to this target
        # point
        k = np.searchsorted(target_points[i] < known_points, True)

        idp = np.arange(k - numsparse // 2, k + (numsparse + 1) // 2)
        if idp[0] < 0:
            idp = np.arange(numsparse)
        if idp[-1] >= m:
            idp = np.arange(m - numsparse, m)
        # xss stores the values from `known_points` used for interpolation on the i'th
        # target point
        xss[i, :] = known_points[idp]
        jdx[i, :] = idp
        idx[i, :] = i

    # Auxiliary vector for computing products of expressions of the form  xss[:,i] - xss[:,j]
    # in order not to include xss[:,i] - xss[:,i] = 0. Will be all ones except for the index i
    # not to include in the running product. this index is updated in the loop
    Iw = np.ones(numsparse, dtype=bool)
    ew = np.zeros((n, 1))
    xtw = np.zeros((n, numsparse - 1))

    Iw[0] = False
    const = np.zeros((n, 1))

    for _ in range(numsparse):
        ew = np.sum(-np.log(np.abs(xss[:, 0].reshape(-1, 1) - xss[:, Iw])), axis=1)
        # normalization constant
        constw = np.exp(ew / numsparse)
        constw = constw.reshape(-1, 1)
        const += constw
    # this normalization constant prevents numerical issues and cancels out in the end
    # not included in return result
    const = const / numsparse

    for j in range(numsparse):
        Iw[j] = False
        # compute the denominator in Eq 3.2 of Berrut and Trefethen
        xtw = const * (xss[:, j].reshape(-1, 1) - xss[:, Iw])
        ws[:, j] = 1 / np.prod(xtw, axis=1)
        Iw[j] = True

    xdiff = xdiff.flatten()
    target_points = target_points.flatten()
    temp = temp.flatten()
    denom = denom.flatten()

    for j in range(numsparse):
        # xdiff[i] is the i'th target point minus the j'th source point for that target pt
        # see the denominator in Eq. 3.3 of Berrut and Trefethen
        xdiff = target_points - xss[:, j]
        temp = ws[:, j] / xdiff
        # vals[:,j] = (1/const)*w_j/(x[i] - xs[j]), with the notation in Eq. 3.3
        vals[:, j] = vals[:, j] + temp
        denom = denom + temp

    # Eq 4.2
    # note that const cancels in numerator and denominator
    vals = vals / denom.reshape(-1, 1)

    vals = vals.flatten()
    idx = idx.flatten()
    jdx = jdx.flatten()
    # A is the linear operator mapping the function values from the fixed source
    # points to the fixed target points.
    # A(i,j) = \ell(x[i] ) w_j/(x[i] - xs[j]), with the notation in Eq. 3.3
    A = sparse.csr_matrix((vals, (idx, jdx)), shape=(n, m), dtype=np.float64)
    A_T = sparse.csr_matrix((vals, (jdx, idx)), shape=(m, n), dtype=np.float64)

    return A, A_T
